#include "GenericC2SPacket.h"
using namespace umcs;

#include "../config.h"
#include <cstring>

CGenericC2SPacket::CGenericC2SPacket(unsigned char *Buffer)
{
    m_Buffer = Buffer;
    m_CurrentByte = 0;
    m_AvailableDataLength = 0;
    m_CurrentField = 0;
}

void CGenericC2SPacket::SetDataLength(unsigned int Length)
{
    m_AvailableDataLength = Length;
    LOG("Packet has %i bytes of data.\n", m_AvailableDataLength);
}

int CGenericC2SPacket::Process(void)
{
    if (GetAvailableBytes() < 1)
    {
        LOG("Available bytes: %i\n", GetAvailableBytes());
        return -1;
    }
    if (m_CurrentByte == 0)
    {
        LOG("Parsing packet ID...\n", 0);
        char Type;
        ReadByte(Type);
        if ((unsigned char)Type != GetPacketID())
        {
            LOG("Incorrect packetID. expexted %02x, got %02x. wat. wat. wat.\n", GetPacketID(), Type);
            return -2;
        }
    }
    for (unsigned int i = m_CurrentField; i < GetNumFields(); i++)
    {
        LOG("Reading field %i.\n", i);
        LOG("FYI, %i available bytes.\n", GetAvailableBytes());
        bool Result = ReadField(i);

        if (Result)
        {
            m_CurrentField++; // field processed, increment current field in case we stop somewhere along the road
            LOG("Field Okay.\n", 0);
        }
        else
        {
            LOG("Not enough bytes (got -1)\n", 0);
            return -1; // field unprocessed - not enough data. stop for now.
        }
    }
    LOG("All fields read.!\n", 0);
    // run the callback for children classes
    AllFieldsRead();
    // be sure to let caller know if some data was unnecessary
    return (m_AvailableDataLength - m_CurrentByte);
}

char CGenericC2SPacket::PeekByte(void)
{
    return (char)m_Buffer[m_CurrentByte];
}

bool CGenericC2SPacket::ReadByte(char &Byte)
{
    if (GetAvailableBytes() < 1)
        return false;

    Byte = PeekByte();
    m_CurrentByte++;
    return true;
}

bool CGenericC2SPacket::ReadInteger(int &Integer)
{
    if (GetAvailableBytes() < 4)
        return false;
    
    ((unsigned char *)(&Integer))[3] = (unsigned char)PeekByte();
    m_CurrentByte++;
    ((unsigned char *)(&Integer))[2] = (unsigned char)PeekByte();
    m_CurrentByte++;
    ((unsigned char *)(&Integer))[1] = (unsigned char)PeekByte();
    m_CurrentByte++;
    ((unsigned char *)(&Integer))[0] = (unsigned char)PeekByte();
    m_CurrentByte++;

    return true;
}

bool CGenericC2SPacket::ReadString(utf8str &String)
{
    // fuck you (notch|java) for making the prefix count the number of characters, not bytes
    if (GetAvailableBytes() < 2)
        return false;

    unsigned short NumChars;
    ((unsigned char *)&NumChars)[1] = m_Buffer[m_CurrentByte];
    ((unsigned char *)&NumChars)[0] = m_Buffer[m_CurrentByte + 1];

    unsigned int StringLength = 0;
    unsigned int NumBytesRead = 2; // 2 for the length

    // proceed to iterate ofer the string for the first time to calculate the number of bytes needed
    for (unsigned short i = 0; i < NumChars; i++)
    {
        if (GetAvailableBytes() < (NumBytesRead + 2))
            return false;
        unsigned short Unit;
        ((unsigned char *)&Unit)[0] = m_Buffer[m_CurrentByte + NumBytesRead];
        ((unsigned char *)&Unit)[1] = m_Buffer[m_CurrentByte + NumBytesRead + 1];
        NumBytesRead += 2;
        StringLength++;

        if (Unit > 0xD7FF && Unit < 0xE000)
        {
            // this is a 2-unit character!
            NumBytesRead += 2;
            StringLength++;
        }
    }
    if (GetAvailableBytes() < NumBytesRead)
        return false;

    // still here? read the string
    utf16str String16;
    for (unsigned int i = 0; i < StringLength; i++)
    {
        unsigned short Unit = *(((unsigned short *)(m_Buffer + m_CurrentByte)) + 1 + i);
        printf("Unit: %c\n", Unit >> 8);
        // endianness swap
        String16.push_back((Unit >> 8) | ((Unit << 8) & 0xFF00));
    }
    UTF16TO8(String16, String);

    m_CurrentByte += NumBytesRead;

    return true;
}